/*
Copyright (C) 2016 Siep Kroonenberg

This file is part of TLaunch.

TLaunch is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TLaunch is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TLaunch.  If not, see <http://www.gnu.org/licenses/>.
*/

/////////////////////////////////////////////////////////
// user options, for now just editor choice.

// The options/editor dialogue is handled here, except that its
// functions are invoked from tlaunch.c and tla_parse.c.

// A globally defined button ed_control has as progid the global variable
// ed_progid, which is set in initialize_editors in this file.

#include "tla_utils.h"
#include "resource.h"
#include "tlaunch.h"

#define MAX_EDITORS 10
// default name below can be overridden by value set in tlaunch.ini
#define CUSTOMED_PROGID L"TL.customed"

static wchar_t * option_class = L"option_class";

/////////////////////////////////////////////////////////
// data structures

typedef struct {
  Progid * progid;
  HWND hw;
  UINT_PTR id;
} Editor;
static Editor editors[MAX_EDITORS + 1]; // one spare for custom editor
static short neds, current_edno, cust_edno;

///////////////////////////////////////////////////////
// a couple of candidates for tla_utils, except that we need them only here

static wchar_t * name_from_cmd( wchar_t * cmd ) {
  wchar_t * prog, * p, * name;
  if( !cmd ) return NULL;
  prog = prog_from_cmd( cmd );
  if( !prog ) {
    return NULL;
  } else if( !prog[0] ) {
    FREE0( prog );
    return NULL;
  }
  p = wcsrchr( prog, L'.' );
  if( p ) *p = L'\0'; // now prog has no extension any more
  p = wcsrchr( prog, L'\\' );
  if( !p ) p = prog;
  else *p = L'\0'; p++;
  name = tl_wcsncpy( p, wcslen( p ));
  name[0] = towupper( name[0] );
  FREE0( prog );
  return name;
} // name_from_cmd

static wchar_t * command_from_shell_cmd( wchar_t * shell_cmd ) {
  int l;
  wchar_t * prog, * command;
  if( !shell_cmd || !shell_cmd[0] ) return NULL;
  // assume shell_cmd was a registry value written by the launcher
  // as the result of the user selecting a custom editor.
  // Just lop off the final bit to get at the command.
  // Otherwise we would have to do a lot more parsing, or
  // store more information.
  prog = prog_from_cmd( shell_cmd );
  // first, existence check
  if( prog && prog[0] && FILE_EXISTS( prog )) {
    l = wcslen( shell_cmd );
    if( l > 5 && !wcscmp( shell_cmd + l - 5, L" \"%1\"" ))
      command = tl_wcsncpy( shell_cmd, l - 5 );
    else if( l > 8 && !wcscmp( shell_cmd + l - 8, L" \"%1\" %*" ))
      command = tl_wcsncpy( shell_cmd, l - 8 );
    else
      command = tl_wcscpy( shell_cmd );
  } else {
    command = NULL;
  }
  FREE0( prog );
  return command;
} // command_from_shell_cmd

///////////////////////////////////////////////////////

// label to be displayed for default editor
wchar_t * def_editorlabel( BOOL wrapped ) {
  if( !ed_progid->name )
    return tl_wcscpy( ed_postfix );
  else
    return tl_wcsnconcat( 3, ed_progid->name, wrapped ? L"\r\n" : L" ",
        ed_postfix );
} // def_editorlabel

// build editors data structure and add custom editor, both as editor and
// as progid; called by parse_ini when it starts with menu- and button specs
void initialize_editors() {

  int i;
  wchar_t * shcmd, * prog, * progi, *icf, *nm;
  wchar_t * exts, * p;
  Progid * pgid;

  // collect editors from progids.
  for( i=0; i<nprogs; i++ ) {
    pgid = progids[i];
    if( !pgid->extensions || !pgid->command ) continue;
    // tex among extensions?
    exts = pgid->extensions;
    p = wcsstr( exts, L".tex" );
    if( p && ( p == exts || *( p - 1 ) == L' ' ) &&
        (*( p + 4 ) == L'\0'  || *( p + 4 ) == L' ' )) {
      editors[neds].progid = pgid;
      // create name if there is none
      if( !pgid->name )
          pgid->name = name_from_cmd( pgid->command );
      if( ++neds >= MAX_EDITORS ) break;
    }
  }
  // add progid and editor for custom editor. Note that
  // the progids and editors arrays each have at least one spare slot
  cust_progid = calloc_fatal( 1, sizeof( Progid ));
  progids[nprogs++] = cust_progid;
  // in case the ini file defined CUSTOMED_PROGID:
  cust_progid->progid = get_env( L"CUSTOMED_PROGID" );
  // otherwise use the #define-d value
  if( !cust_progid->progid ) cust_progid->progid = CUSTOMED_PROGID;
  cust_progid->extensions = L".tex .sty .cls";
  cust_edno = neds++;
  editors[cust_edno].progid = cust_progid;

  // a working cust_progid in the registry gets priority
  get_assoc_data( cust_progid->progid, NULL, &cust_progid->shell_cmd,
      &cust_progid->iconspec, &cust_progid->name );
  // command_from_shell_cmd also checks for existence of the program
  cust_progid->command = command_from_shell_cmd( cust_progid->shell_cmd );
  if( !cust_progid->command ) {
    FREE0( cust_progid->shell_cmd );
  } else if( !cust_progid->name ) {
    cust_progid->name = name_from_cmd( cust_progid->command );
  }

  // currently configured editor
  current_edno = -1;
  get_assoc_data( L".tex", NULL, &shcmd, &icf, &nm );
  if( shcmd ) {
    prog = prog_from_cmd( shcmd );
    if( FILE_EXISTS( prog )) {
      // is this a configured editor?
      for( i=0; i<neds; i++ ) {
        pgid = editors[i].progid;
        if( !pgid->command ) continue; // might only happen with cust_edno
        progi = prog_from_cmd( pgid->command );
        if( same_file( prog, progi )) {
          current_edno = i;
          FREE0( progi );
          FREE0( shcmd );
          FREE0( icf );
          break;
        } else {
          FREE0( progi );
        }
      } // end for
      if( current_edno < 0 ) { // prog exists, but not as configured editor
        current_edno = cust_edno;
        cust_progid->shell_cmd = shcmd;
        cust_progid->command = command_from_shell_cmd( shcmd );
        // use specified icon, or create default
        if( icf ) cust_progid->iconspec = icf;
        else cust_progid->iconspec = tl_wcsnconcat( 2, prog, L",0" );
        if( nm ) cust_progid->name = nm;
        else cust_progid->name = name_from_cmd( shcmd );
      }
      FREE0( prog );
    } else { // prog does not exist: ignore
      FREE0( prog );
      FREE0( shcmd );
      FREE0( icf );
      current_edno = 0;
    }
  } else { // nothing useful configured
    FREE0( icf );
    current_edno = 0;
  }
  ed_progid = editors[current_edno].progid; // may be an unconfigured cust_progid
  ed_initdone = TRUE;
} // initialize_editors

/////////////////////////////////////////////////////////
// building interface

static HWND add_options_child( wchar_t * cls, wchar_t * text, DWORD style,
    DWORD x, DWORD y, DWORD w, DWORD h, HWND parent, DWORD id ) {
  HWND hC;
  hC = CreateWindow( cls, text,
      WS_CHILD | WS_VISIBLE | style, x, y, w, h,
      parent, (HMENU)id, GetModuleHandle( NULL ), NULL );
  if( !( hC )) tldie( parent, L"Failure to create child" );
  SendMessage( hC, WM_SETFONT, ( WPARAM )hfnt, TRUE );
  return hC;
}

void update_ed_control() {
  wchar_t * s;
  UINT ctrl_id;

  if( !ed_control ) return;
  ed_control->enabled = ed_progid->command ? TRUE : FALSE;
  EnableWindow( GetDlgItem( hMain, ed_control->id ), ed_control->enabled );

  s = def_editorlabel( ed_isbutton );
  ctrl_id = ed_control->id;
  if( ed_isbutton ) ctrl_id++;
  SendDlgItemMessage( hMain, ctrl_id , WM_SETTEXT, 0, (LPARAM)s );
  FREE0( s );

  if( ed_isbutton ) {
    get_icon( ed_control );
    SendMessage( ed_control->hw, BM_SETIMAGE,
        ( WPARAM )IMAGE_ICON, ( LPARAM )ed_control->hicon );
  }
}

// update widgets of this window i.e. radio choice and cutom editor,
// and also the editor button plus label of the main window inasfar as
// they exist
static void update_ed_widgets( HWND hwnd ) {
  int i;
  Progid * pgid;
  if( neds > 1 ) {
    for( i=0; i < neds; i++ ) {
      pgid = editors[i].progid;
      if( pgid->name )
        SetWindowText( editors[i].hw, pgid->name );
      else if( i==cust_edno )
        SetWindowText( editors[i].hw, L"(custom editor)" );
      if( i == current_edno ) {
        SendMessage( editors[i].hw, BM_SETCHECK, 1, 0 );
        ed_progid = pgid;
        if( ed_control ) ed_control->progid = pgid;
      } else {
        SendMessage( editors[i].hw, BM_SETCHECK, 0, 0 );
      }
    } // end for
    if( cust_progid->command ) {
      EnableWindow( editors[cust_edno].hw, TRUE );
      ShowWindow( editors[cust_edno].hw, SW_SHOWNA );
    } else {
      EnableWindow( editors[cust_edno].hw, FALSE );
      ShowWindow( editors[cust_edno].hw, SW_HIDE );
    }
  } // neds > 1
  update_ed_control();
} // update_ed_widgets

static BOOL pick_custom_ed( HWND hwnd ) {
  OPENFILENAME ofn;
  BOOL chosen;
  ZeroMemory( &ofn, sizeof( OPENFILENAME ));
  ofn.lStructSize = sizeof( OPENFILENAME );
  ofn.hwndOwner = hwnd;
  ofn.lpstrFilter =
      L"Programs (*.exe, *.bat, *.cmd)\0*.exe;*.bat;*.cmd)\0\0";
  ofn.nFilterIndex = 1;
  ofn.lpstrFile = calloc_fatal( MAX_PATH, sizeof( wchar_t ));
  ofn.nMaxFile = MAX_PATH;
  ofn.lpstrTitle = L"Select custom editor";
  ofn.Flags = OFN_FILEMUSTEXIST;
  ofn.lpstrDefExt = L"exe";
  chosen = GetOpenFileName( &ofn );
  if( chosen ) {
    FREE0( cust_progid->command );
    FREE0( cust_progid->shell_cmd );
    FREE0( cust_progid->name );
    FREE0( cust_progid->iconspec );
    cust_progid->iconspec = tl_wcsnconcat( 2, ofn.lpstrFile, L",0" );
    cust_progid->command = tl_wcsnconcat( 3, L"\"", ofn.lpstrFile, L"\"" );
    cust_progid->shell_cmd = tl_wcsnconcat( 2, cust_progid->command,
        L" \"%1\"" );
    cust_progid->name = name_from_cmd( cust_progid->command );
    current_edno = cust_edno;
    update_ed_widgets( hwnd );
  }
  FREE0( ofn.lpstrFile );
  return chosen;
}

void options_tl( HWND hwnd ) {
  HWND hOpt;
  hOpt = CreateWindowEx(
      WS_EX_TOPMOST,
      option_class,
      L"Editor Choice",
      WS_POPUP | WS_OVERLAPPED | WS_CAPTION | WS_THICKFRAME |
          WS_VISIBLE,
      CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
      hMain, NULL, NULL, NULL );
  if ( hOpt ) {
    ShowWindow( hOpt, SW_SHOW);
    UpdateWindow( hOpt);
    EnableWindow( hwnd, FALSE );
  } else {
    tlwarn( hwnd, L"Creation Editor Window Failed" );
  }
}

LRESULT CALLBACK OptProc( HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam ) {

  int i, j;
  HWND hClose, hCustom;
  UINT_PTR id_next;
  Progid * pgid;
  switch( msg ) {
    case WM_CREATE: {

      int x, y, w, h;
      RECT rect = { -1, -1, -1, -1 };

      if( GetWindowRect( hMain, &rect ))
        SetWindowPos( hwnd, HWND_TOP, rect.left + 30, rect.top + 30,
            30 * cx, ( neds>1 ) ? (( neds + 4 ) * 2 * cy ) : ( 8 * cy ),
            SWP_NOACTIVATE | SWP_NOZORDER );
      else
        SetWindowPos( hwnd, HWND_TOP, 0, 0,
            30 * cx, ( neds>1 ) ? (( neds + 5 ) * 2 * cy ) : ( 9 * cy ),
            SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOZORDER );
      /*
      create some radio buttons
      CreateWindow: may use last parameter for a pointer to
      a data structure; see lpCreateParams member of CREATESTRUCT structure
      */

      x = 2 * cx;
      y = cy / 2;
      w = 24 * cx;
      h = 2 * cy;
      id_next = ID_FIRST_OPT_CHILD;
      if( neds > 1 ) {
        for( i=0; i<neds; i++ ) {
          editors[i].id = id_next++;
          editors[i].hw = add_options_child( L"button", NULL, BS_RADIOBUTTON,
              x, y, w, h, hwnd, editors[i].id );
          y += 2 * cy;
        }
        if( !cust_progid || !cust_progid->command ) {
          ShowWindow( editors[cust_edno].hw, SW_HIDE );
          EnableWindow( editors[cust_edno].hw, FALSE );
        }
      }

      // custom selection button
      hCustom = add_options_child( L"button", L"Custom editor...",
          BS_PUSHBUTTON | WS_TABSTOP, x, y, w, h, hwnd, ID_CUSTOM );
      if( !hCustom ) tldie( hwnd, L"Failure to create button" );
      y += 5 * cy / 2;

      // close button
      hClose = add_options_child( L"button", L"Close",
          BS_PUSHBUTTON | WS_TABSTOP, x, y, w, h, hwnd, ID_CLOSE_OPT );
      if( !hClose ) tldie( hwnd, L"Failure to create button" );

      update_ed_widgets( hwnd );
      return 0;
    } // end WM_CREATE

    case WM_CLOSE:
      EnableWindow( hMain, TRUE );
      DestroyWindow( hwnd );
      return 0;

    case WM_COMMAND: {
      DWORD id = LOWORD( wParam );
      // editor radio button?
      if( id >= editors[0].id && id <= editors[ neds - 1 ].id ) {
        for( j=0; j<neds; j++ ) if( id == editors[j].id ) {
          pgid = editors[j].progid;
          if( pgid->command ) {
            current_edno = j;
            update_ed_widgets( hwnd );
          } // if pgid->command
        } // for j if id
        return 0;
      } else if( id == ID_CUSTOM ) {
        // file browser
        if( !pick_custom_ed( hwnd )) return 0;
        if( cust_progid->command ) {
          current_edno = cust_edno;
          if( editors[cust_edno].hw ) {
            ShowWindow( editors[cust_edno].hw, SW_SHOW );
            EnableWindow( editors[cust_edno].hw, TRUE );
          }
        }
        update_ed_widgets( hwnd );
        return 0;
      } else if( id == ID_CLOSE_OPT ) {
        // GUI updates already done
        set_assoc_abs( L".tex", ed_progid->progid );
        set_assoc_abs( L".sty", ed_progid->progid );
        set_assoc_abs( L".cls", ed_progid->progid );
        register_progid( cust_progid->progid, cust_progid->shell_cmd,
            cust_progid->iconspec, cust_progid->name );
        PostMessage( hwnd, WM_CLOSE, 0, 0 );
        return 0;
      }
    } // end WM_COMMAND
    default: {
      return DefWindowProc( hwnd, msg, wParam, lParam );
    }
  } // switch msg
} // OptProc

void register_opt_wnd_class( ) {
  WNDCLASSEX woptionsc;

  woptionsc.cbSize = sizeof(WNDCLASSEX);
  woptionsc.style = 0; // set by CreateWindow
  woptionsc.lpfnWndProc = OptProc;
  woptionsc.cbClsExtra = 0;
  woptionsc.cbWndExtra = 0;
  woptionsc.hInstance = hInst;
  woptionsc.hIcon =
      LoadIcon( GetModuleHandle( NULL ), MAKEINTRESOURCE( IDI_TL ));
  woptionsc.hCursor = NULL;
  woptionsc.hbrBackground = ( HBRUSH )( COLOR_BTNFACE + 1 );
  woptionsc.lpszMenuName = NULL;
  woptionsc.lpszClassName = option_class;
  woptionsc.hIconSm = NULL;

  if(!RegisterClassEx( &woptionsc ))
      tldie( NULL, L"Options Window Registration Failed!" );
}
