/*
Copyright (C) 2016 Siep Kroonenberg

This file is part of TLaunch.

TLaunch is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TLaunch is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TLaunch.  If not, see <http://www.gnu.org/licenses/>.
*/

// Main program: launcher for TeX Live.
// First-time initialization includes adjustment of user searchpath,
// setting up file associations and creating an uninstaller/un-initializer.

#include "tla_utils.h"
#include "tlaunch.h"

// The code below is the easy way to initialize an array of structs and
// its length, but cannot be used in a shared header file, because then
// these variables would be multiply defined. Therefore,
// wchar_t * perl_envs is declared in tlaunch.h and set to perl_envs_arr
// in set_simple_globals, where its length nperl_envs is also calculated.

PerlEnv perl_envs_arr [] = {
  { L"PERL5OPT", NULL },
  { L"PERLIO", NULL },
  { L"PERLIO_DEBUG", NULL },
  { L"PERLLIB", NULL },
  { L"PERL5DB", NULL },
  { L"PERL5DB_THREADED", NULL },
  { L"PERL5SHELL", NULL },
  { L"PERL_ALLOW_NON_IFS_LSP", NULL },
  { L"PERL_DEBUG_MSTATS", NULL },
  { L"PERL_DESTRUCT_LEVEL", NULL },
  { L"PERL_DL_NONLAZY", NULL },
  { L"PERL_ENCODING", NULL },
  { L"PERL_HASH_SEED", NULL },
  { L"PERL_HASH_SEED_DEBUG", NULL },
  { L"PERL_ROOT", NULL },
  { L"PERL_SIGNALS", NULL },
  { L"PERL_UNICODE", NULL },
};
// the val members of the array elements will be retrieved right after
// processing the ini file

// note that globals without explicit initialization
// are already automatically set to 0 or NULL by the compiler
// so some of the initializations below are no-ops.

void set_simple_globals() {

  // basename of uninstaller
  uninstbase = L"tl_undo";

  // filetype policy. 0: do nothing; 1: do not overwrite; 2: always
  assoc_mode = 1;
  do_searchpath = 1;

  ed_initdone = FALSE;

  perl_envs = perl_envs_arr;
  nperl_envs = ((int)( sizeof( perl_envs_arr ) / sizeof( perl_envs[0] )));

  // default editor
  ed_control = NULL;
  ed_progid = NULL;
  ed_isbutton = FALSE;

  // announcement string
  tl_announce = NULL;
}

// directory for the uninstaller start menu item
wchar_t * get_uninst_menu() {
  wchar_t * menuprefix, * fullpath;

  if( !tl_name ) return NULL;
  // this should not happen, since the parser supplies a default name

  menuprefix = get_shell_folder( FOLDERID_Programs );
  if( !menuprefix || !( *menuprefix )) return NULL;
  fullpath = tl_wcsnconcat( 3, menuprefix, L"\\", tl_name );
  FREE0( menuprefix );
  return fullpath;
} // get_uninst_menu

wchar_t * kpse_answer( wchar_t * question ) {
  wchar_t * command, * out_buf;
  command = tl_wcsnconcat( 4, L"\"", root,
        L"\\bin\\win32\\kpsewhich.exe\" ", question );
  out_buf = capture_output( command, FALSE );
  FREE0( command );
  if( out_buf && !out_buf[0] ) FREE0( out_buf );
  return out_buf;
} // kpse_answer

// first-time initialization: path, file associations, uninstaller

// Note. Re-initialization consists of uninitializing and restarting.
void first_time( HWND hwnd, wchar_t ** mess ) {
  wchar_t * str = NULL, * ext, * exts, * p;
  int i, exts_len;
  *mess = NULL;

  // run pre-inst batchfile, if any
  if( pre_inst && FILE_EXISTS( pre_inst )) {
    str = capture_tmp( pre_inst, TRUE );
    if( str ) {
      add_to_mess( mess, str );
      FREE0( str );
    }
  }

  // user path
  if( do_searchpath ) {
    if( !add_to_user_path( tlbin )) {
      add_to_mess( mess, L"Failed to add %ls to user searchpath", tlbin );
    }
  }

  // file associations
  if( assoc_mode ) {
    for( i=0; i<nprogs; i++ ) {
      // register progid
      if( !register_progid( progids[i]->progid, progids[i]->shell_cmd,
          progids[i]->iconspec, progids[i]->name )) {
        add_to_mess( mess,
            L"Failed to register filetype %ls", progids[i]->progid );
        continue;
      }
      // loop through all its extensions for primary and secondary associations
      if( progids[i]->extensions ) {
        exts_len = wcslen( progids[i]->extensions );
        exts = tl_wcscpy( progids[i]->extensions );
        ext = exts;
        while( ext && *ext ) {
          // find next separator/space
          p = wcschr( ext, L' ' );
          if( p ) *p = L'\0'; // now ext is 1 extension
          if( !set_assoc( ext, progids[i]->progid, assoc_mode ))
              add_to_mess( mess, L"Failed to associate %ls with filetype %ls",
              ext, progids[i]->progid );
          if( !p ) break;
          ext = ( p < ( exts + exts_len )) ? p + 1 : NULL;
        } // end running through extensions
        FREE0( exts );
      }
    } // end running through progids
  }

  // create uninstaller: copy launcher and ini file to uninstdir/uninst.[exe|ini]
  {
    wchar_t * uninst_file = NULL, * configname, * str2;
    BYTE * buf;
    __int64 exesize; // is signed
    HANDLE h;
    DWORD written, lread;
    int ok = 1; // valid uninstaller?
    if( !uninstdir ) {
      add_to_mess( mess,
          L"No uninstall directory specified, therefore no uninstaller" );
      goto end_uninst; // not an error
    }
    if( !mkdir_recursive( uninstdir )) {
      add_to_mess( mess, L"Failed to create %ls", uninstdir );
      ok = 0;
      goto end_uninst;
    }
    texmfvar = kpse_answer(  L"-var-value=TEXMFVAR" );
    texmfconfig = kpse_answer(  L"-var-value=TEXMFCONFIG" );
    // used in uninstall; failure not fatal
    uninst_file = tl_wcsnconcat( 4, uninstdir, L"\\", uninstbase, L".ini" );

    // Parameters CreateFile for creating or overwriting a file:
    // 1. filename
    // 2. desired access: GENERIC_READ | GENERIC_WRITE
    // 3. share mode: in the current case, 0 (exclusive access) is ok.
    //    this access mode holds while the handle is open.
    // 4. security attributes. this is the reason why we do not use CopyFile
    //    which copies security attributes; we just want
    //    default attributes. So we set this parameter NULL.
    // 5. CreationDisposition, set to CREATE_ALWAYS i.e. overwrite
    //    if already existing
    // 6. flags; FILE_ATTRIBUTE_NORMAL will do
    // 7. optional template handle; set to NULL
    // Return value is a file handle or otherwise INVALID_HANDLE_VALUE

    h = CreateFile( uninst_file, GENERIC_READ|GENERIC_WRITE, 0,
        NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
    if( h == INVALID_HANDLE_VALUE ) {
      add_to_mess( mess, L"Failed to create %ls", uninst_file );
      h = NULL;
      ok = 0;
      goto end_uninst;
    }
    // first write root section, with definition for root
    str = tl_wcsnconcat( 3, L"[Root]\r\nroot=", root, L"\r\n" );
    // note: write strings without terminating 0
    ok = WriteFile( h, (void *)str, wcslen( str ) * sizeof( wchar_t ),
          &written, NULL );
    FREE0( str );
    if( !ok ) {
      CloseHandle( h );
      h = NULL;
      add_to_mess( mess, L"Problem writing %ls", uninst_file );
      goto end_uninst;
    }
    // add VERSION, and TEXMFVAR and TEXMFCONFIG if applicable
    str = tl_wcsnconcat( 3, L"version=", version, L"\r\n" );
    ok = WriteFile( h, (void *)str, wcslen( str ) * sizeof( wchar_t ),
        &written, NULL );
    FREE0( str );
    if( !ok ) {
      CloseHandle( h );
      h = NULL;
      add_to_mess( mess, L"Problem writing %ls", uninst_file );
      goto end_uninst;
    }
    if( texmfvar ) {
      str = tl_wcsnconcat( 3, L"texmfvar=", texmfvar, L"\r\n" );
      ok = WriteFile( h, (void *)str, wcslen( str ) * sizeof( wchar_t ),
          &written, NULL );
      FREE0( str );
      if( !ok ) {
        CloseHandle( h );
        h = NULL;
        add_to_mess( mess, L"Problem writing %ls", uninst_file );
        goto end_uninst;
      }
    }
    if( texmfconfig ) {
      str = tl_wcsnconcat( 3, L"texmfconfig=", texmfconfig, L"\r\n" );
      ok = WriteFile( h, (void *)str, wcslen( str ) * sizeof( wchar_t ),
          &written, NULL );
      FREE0( str );
      if( !ok ) {
        CloseHandle( h );
        h = NULL;
        add_to_mess( mess, L"Problem writing %ls", uninst_file );
        goto end_uninst;
      }
    }
    // append ini file wholesale
    configname = tl_wcsnconcat( 4, root, L"\\", tlaunchbase, L".ini" );
    if( !FILE_EXISTS( configname )) {
      FREE0( configname );
      str = tl_wcsnconcat( 3, L"--format=texmfscripts ", tlaunchbase, L".ini" );
      configname = kpse_answer( str );
    }
    if( !configname )
        // this shouldn't happen since the launcher just read this file
        tldie( hwnd, L"Config file %ls.ini not found", tlaunchbase );
    str = slurp_file_to_wcs( configname, MAX_CONFIG );
    ok = WriteFile( h, (void *)str, wcslen( str ) * sizeof(wchar_t ),
        &written, NULL );
    FREE0( str );
    CloseHandle( h );
    h = NULL;
    FREE0( configname );
    if( !ok ) {
      add_to_mess( mess, L"Problem writing %ls", uninst_file );
      goto end_uninst;
    }
    // done with uninst ini file; now copy the launcher executable.
    // first, copy exe to in-memory buffer buf
    exesize = win_filesize( _wpgmptr );
    buf = slurp_file( _wpgmptr, exesize, &lread );
    // launcher binary read to buf; now write buf
    // just need to change extension of uninst_file
    wcscpy( uninst_file + wcslen( uninst_file ) - 3, L"exe" );
    h = CreateFile( uninst_file, GENERIC_READ|GENERIC_WRITE, 0,
        NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
    if( h == INVALID_HANDLE_VALUE ) {
      add_to_mess( mess, L"Failed to create %ls", uninst_file );
      ok = 0;
      goto end_uninst;
    }
    ok = WriteFile( h, buf, lread, &written, NULL );
    FREE0( buf );
    CloseHandle( h );
    if( !ok ) {
      add_to_mess( mess, L"Cannot copy launcher" );
      goto end_uninst;
    }

    // create shortcut to uninstaller
    // on failure, do NOT set ok to 0; we want to keep the uninstaller itself
    str = get_uninst_menu();
    if( !mkdir_recursive( str )) {
      add_to_mess( mess, L"Cannot create shortcut for uninstaller (%ls)",
          uninst_file );
      goto end_uninst;
    }
    str2 = tl_wcsnconcat( 2, str, L"\\Remove TeX Live settings.lnk" );
    if( !make_shortcut( uninst_file, str2 )) {
      add_to_mess( mess, L"Cannot create shortcut for uninstaller (%ls)",
          uninst_file );
      rmdir_recursive( str );
      goto end_uninst;
    }
    FREE0( str2 );
    // create registry entry for uninstaller, to convince windows
    // that it was a proper and successful installation
    // no error checking
    {
      wchar_t * keypath;
      keypath = tl_wcsnconcat( 2,
          L"software\\microsoft\\windows\\currentversion\\uninstall\\",
          tl_name );
      reg_set_stringvalue( HKEY_CURRENT_USER, keypath, L"DisplayName",
          tl_name );
      reg_set_stringvalue( HKEY_CURRENT_USER, keypath, L"DisplayVersion",
          version );
      str = tl_wcsnconcat( 3, L"\"", uninst_file, L"\"" );
      reg_set_stringvalue( HKEY_CURRENT_USER, keypath, L"UninstallString",
          str );
      FREE0( str );
      FREE0( keypath );
    }
    // cleanup: uninst files? uninstdir?
    end_uninst:;
    if( uninst_file ) FREE0( uninst_file );
    if( !ok && uninstdir && ( FILE_EXISTS( uninstdir )))
      rmdir_recursive( uninstdir );
  } // done creating uninstaller

  // run post-inst batchfile, if any
  if( post_inst && FILE_EXISTS( post_inst )) {
    str = capture_tmp( post_inst, TRUE );
    if( str ) {
      add_to_mess( mess, str );
      FREE0( str );
    }
  }
} // first_time

// called by clear_tl
void uninstall( wchar_t ** mess ) {
  wchar_t * ext, * exts, * p, * str;
  int i, exts_len;

  // run pre-uninst batchfile, if any
  if( pre_uninst && FILE_EXISTS( pre_uninst )) {
    add_to_mess( mess, L"Found pre-uninst" );
    str = capture_tmp( pre_uninst, TRUE );
    if( str ) {
      add_to_mess( mess, str );
      FREE0( str );
    }
  }

  // file associations
  for( i=0; i<nprogs; i++ ) {
    // loop through all extensions for primary and secondary associations
    if( progids[i]->extensions ) {
      exts_len = wcslen( progids[i]->extensions );
      exts = tl_wcscpy( progids[i]->extensions );
      ext = exts;
      while( ext && *ext ) {
        // find next separator/space
        p = wcschr( ext, L' ' );
        if( p ) *p = L'\0'; // now ext is 1 extension
        if( !remove_assoc( ext, progids[i]->progid ))
          add_to_mess( mess,
            L"Failed to remove association %ls with filetype %ls",
            ext, progids[i]->progid );
        if( !p ) break;
        ext = ( p < ( exts + exts_len )) ? p + 1 : NULL;
      } // end running through extensions
      FREE0( exts );
    }
    // unregister progid
    if( !remove_progid( progids[i]->progid )) {
      add_to_mess( mess,
          L"Failed to remove filetype %ls", progids[i]->progid );
      continue;
    }
  } // end running through progids

  add_to_mess( mess, L"Finished clearing file associations" );

  // texmfvar and texmfconfig and their parent, if empty.
  // kpse_answer will return NULL rather than an empty string.

  if( !texmfvar ) texmfvar = kpse_answer( L"-var-value=TEXMFVAR" );
  if( texmfvar ) {
    wchar_subst( texmfvar, L'/', L'\\' );
    if( FILE_EXISTS( texmfvar )) rmdir_recursive( texmfvar );
    if( FILE_EXISTS( texmfvar ))
      add_to_mess( mess, L"Failed to delete texmfvar: %ls", texmfvar );
  }
  if( !texmfconfig ) texmfconfig = kpse_answer( L"-var-value=TEXMFCONFIG" );
  if( texmfconfig ) {
    wchar_subst( texmfconfig, L'/', L'\\' );
    if( FILE_EXISTS( texmfconfig )) rmdir_recursive( texmfconfig );
    if( FILE_EXISTS( texmfconfig ))
      add_to_mess( mess, L"Failed to delete texmfconfig: %ls", texmfconfig );
  }
  if( texmfvar ) {
    // remove parent if it exists and is empty
    p = wcsrchr( texmfvar, L'\\' );
    if( p && p - texmfvar > 0 ) {
      DWORD last_err;
      *p = L'\0';
      // RemoveDirectory only removes non-empty directories
      if( FILE_EXISTS( texmfvar )) {
        if( !RemoveDirectory( texmfvar )) {
          last_err = GetLastError();
          if( last_err == ERROR_DIR_NOT_EMPTY )
            add_to_mess( mess, L"%ls not removed because non-empty",
                texmfvar );
          else
            add_to_mess( mess, L"%ls not removed because of error %ul",
                texmfvar, last_err );
        }
      }
    }
  }

  // user path
  if( !remove_from_user_path( tlbin ))
    add_to_mess( mess, L"Failed to remove %ls from user searchpath", tlbin );

  // remove uninstaller shortcut and its subdirectory;
  // leave removing uninstaller to caller, i.e. uninstall-self or clear_tl
  str = get_uninst_menu();
  rmdir_recursive( str );
  if( FILE_EXISTS( str ))
    add_to_mess( mess, L"Failed to remove uninstaller shortcut %ls", str );
  FREE0( str );
  str = tl_wcsnconcat( 2,
      L"software\\microsoft\\windows\\currentversion\\uninstall\\",
      tl_name );
  if( !reg_del_key( HKEY_CURRENT_USER, str ))
    add_to_mess( mess, L"Failed to remove uninstaller registry entry" );
  FREE0( str );
  // the caller should handle any accumulated messages and terminate
} //uninstall

// used when program is called as uninstaller
void uninstall_self( ) {
  wchar_t * cmd, * comspec, * mess = NULL;

  uninstall( &mess );
  // can't come back to report; instead, display warnings so far now
  if( mess ) {
    tlinfo( NULL, mess );
    FREE0( mess );
  }
  // !!! hack alert !!!
  comspec = get_env( L"COMSPEC" );
  cmd = tl_wcsnconcat( 3, L" /c \"timeout /t 3 /nobreak >NUL && rmdir /s /q \"",
      uninstdir, L"\" >NUL\"" );
  ShellExecute( NULL, NULL, comspec, cmd, NULL, SW_HIDE );
  ExitProcess( GetLastError() );
} // uninstall-self

////////////////////////////////////////////////
// the _tl functions

// terminate the launcher normally
void quit_tl( HWND hwnd ) {
  PostMessage( hwnd, WM_CLOSE, 0, 0 );
}

// TODO: replace with something prettier
void about_tl( HWND hwnd ) {
  wchar_t * mess;
  #include "versioninfo.h"
  tl_wasprintf( &mess, L"TLaunch %ls\n\nCopyright (C) %ls, %ls\n\nWebpage %ls",
      VERSION, AUTHOR, COPYRIGHT, URL );
  MessageBox( hwnd, mess, tl_name, 0 );
  FREE0( mess );
}

// initialization: clear settings and restart
void init_tl( HWND hwnd ) {
  wchar_t * mess = NULL;
  wchar_t * prog;
  // wchar_t * buf;
  PROCESS_INFORMATION pi;
  STARTUPINFO si;

  uninstall( &mess );
  // remove uninstaller directory
  rmdir_recursive( uninstdir );
  if( FILE_EXISTS( uninstdir ))
    add_to_mess( &mess, L"Failed to remove %ls with uninstaller", uninstdir );
  add_to_mess( &mess, L"If necessary, restart manually." );
  display_string( hwnd, mess, L"Re-initialize" );
  FREE0( mess );

  prog = tl_wcscpy( _wpgmptr );
  /* alternative:
    buf = calloc_fatal( TL_MAX_STR, sizeof( wchar_t ));
    // #define TL_MAX_STR 32768 is unicode max_path plus one
    ZeroMemory( buf, TL_MAX_STR * sizeof( wchar_t ));
    if( GetModuleFileName( NULL, buf, TL_MAX_STR ) >= TL_MAX_STR || !buf[0] )
        ExitProcess( GetLastError() );
    prog = tl_wcscpy( buf );
    FREE0( buf );
  */
  ZeroMemory( &pi, sizeof( pi ));
  ZeroMemory( &si, sizeof( si ));
  si.cb = sizeof(STARTUPINFO);
  if( !CreateProcess( prog, NULL, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi ))
      tlwarn( NULL, L"Failure to restart; do this manually" );
  PostQuitMessage( 0 );
} // init_tl

// undo initializations
void clear_tl( HWND hwnd ) {
  wchar_t * mess = NULL;

  uninstall( &mess );
  // remove uninstaller directory
  rmdir_recursive( uninstdir );
  if( FILE_EXISTS( uninstdir ))
    add_to_mess( &mess, L"Failed to remove %ls with uninstaller", uninstdir );
  add_to_mess( &mess, L"Done with cleanup." );
  display_string( hwnd, mess, L"Uninstall" );
  FREE0( mess );
  PostQuitMessage( 0 );
} // clear_tl

void clean_perl_env() {
  int i;
  SetEnvironmentVariable( L"PATH", tlp_searchpath );
  for( i=0; i<nperl_envs; i++ )
    SetEnvironmentVariable( perl_envs[i].name, NULL );
}

void restore_perl_env() {
  int i;
  SetEnvironmentVariable( L"PATH", new_searchpath );
  for( i=0; i<nperl_envs; i++ )
    SetEnvironmentVariable( perl_envs[i].name, perl_envs[i].val );
}

// initializations to do before parsing the ini file
void set_installer_globals( wchar_t ** configname ) {
  // Run as launcher if kpsewhich is in the same directory as the program
  // or in its bin/win32 subdirectory. Otherwise, run as uninstaller.
  // In the uninstaller case, the ini file should be in the same directory.
  // This function also sets either:
  // - if launcher, the global variable uninstdir. Otherwise
  // - the globals root and tlaunchbase i.e. basename of launcher

  wchar_t * exepath, * p, * kp = NULL, * cmd = NULL;

  exepath = tl_wcscpy( _wpgmptr ); // need copy; exepath will be modified

  if( _wcsicmp( exepath + wcslen( exepath ) - 4, L".exe" ))
      tldie( NULL, L"Unexpected filename: %ls", exepath );
  // the inifile should have the name of the launcher or uninstaller,
  // but with extension .ini
  *configname = tl_wcscpy( exepath );
  wcscpy( *configname + wcslen( *configname ) - 3, L"ini" );
  p = wcsrchr( exepath, L'\\' );
  if( !p ) tldie( NULL, L"Own path %ls not an absolute filename", exepath );
  *p = L'\0';
  // now exepath is directory; p+1 is basename
  if( FILE_EXISTS( *configname )) {
    kp = tl_wcsnconcat( 2, exepath, L"\\bin\\win32\\kpsewhich.exe" );
    if( FILE_EXISTS( kp )) { // we are launcher, in installation root
      is_uninstaller = 0;
      root = tl_wcscpy( exepath );
      p++; // start of basename
      tlaunchbase = tl_wcsncpy( p, wcslen( p ) - 4 );
    } else {
      is_uninstaller = 1;
      uninstdir = tl_wcscpy( exepath );
    }
    FREE0( kp );
  } else { // launcher in bin/win32, config found by kpsewhich
    FREE0( *configname ); // get rid of wrong guess
    kp = tl_wcsnconcat( 2, exepath, L"\\kpsewhich.exe" );
    if( !FILE_EXISTS( kp ))
        tldie( NULL, L"The launcher should be either in the root of TeX Live "
        "or in the bin\\win32 subdirectory" );
    p++;
    tlaunchbase = tl_wcsncpy( p, wcslen( p ) - 4 );
    root = tl_wcsncpy( exepath, wcslen( exepath ) - wcslen( L"\\bin\\win32" ));
    cmd = tl_wcsnconcat( 3, L"--format=texmfscripts ",
        tlaunchbase, L".ini" );
    *configname = kpse_answer( cmd );
    if( !( *configname )) tldie( NULL, L"Config file %ls not found", tlaunchbase );
    FREE0( cmd );
    FREE0( kp );
  }
  if( !is_uninstaller )
  FREE0( exepath );
} // set_installer_globals

void get_config( HWND hwnd ) {

  wchar_t * config, * configname, *mess = NULL, * uninstname;
  int do_init;

  set_installer_globals( &configname );
  config = slurp_file_to_wcs( configname, MAX_CONFIG );
  parse_ini( config );
  FREE0( config ); // don't save; when needed, get fresh, unmangled copy
  FREE0( configname );

  do_init = 0;
  if( !uninstdir )
      tldie( hwnd, L"Should not happen: no uninstaller folder defined" );
  uninstname = tl_wcsnconcat( 4, uninstdir, L"\\", uninstbase, L".ini" );
  if( !FILE_EXISTS( uninstname )) {
    do_init = 1;
  } else {
    wcscpy( uninstname + wcslen( uninstname ) - 3, L"exe" );
    if( !FILE_EXISTS( uninstname )) do_init = 1;
  }
  FREE0( uninstname );

  if( do_init ) {
    HWND hsplash;
    hsplash = splash( NULL, L"Initializing..." );
    first_time( hwnd, &mess ); // path, file associations, uninstaller
    SendMessage( hsplash, WM_CLOSE, 0, 0 );
    add_to_mess( &mess, L"Done initializing" );
    display_string( NULL, mess, L"Initializations" );
    // wait a while to allow build_gui to see the changed registry settings
    // Sleep( 100 );
    FREE0( mess );
  }
} // get_config

/////////////////////////////////////////

BOOL run_control( HWND hwnd, UINT_PTR id ) {
  Control * ctrl = NULL;
  int i;
  for( i=0; i<ncontrols; i++ ) {
    if( controls[i] && ( controls[i]->id == id )) {
      ctrl = controls[i];
      break;
    }
  }
  if( !ctrl ) return FALSE; // i.e. message not handled
  if( ctrl->enabled ) {
    switch( ctrl->type ) {
      case L'S': {
        HWND hsplash;
        wchar_t * outp;
        Script * prs = ctrl->script;
        // existence of command and splashtext members is guaranteed
        hsplash = splash( hwnd, prs->splashtext );
        outp = capture_output( prs->command, TRUE );
        SendMessage( hsplash, WM_CLOSE, 0, 0 );
        if( !outp )
          tlwarn( hwnd, L"No output from \n", prs->command );
         else {
           display_string( hwnd, outp, prs->command ); // modal dialog
          FREE0( outp );
        }
        return TRUE;
      }
      case L'D': {
        int l;
        l = (int)ShellExecute( NULL, NULL, (wchar_t *)(ctrl->action),
            NULL, NULL, SW_SHOWNORMAL ); // (int) cast as per MS docu
        if( l <= 32 )
          tlwarn( hwnd, L"Failure to open %ls, error %d",
              (wchar_t *)(ctrl->action), l );
        // A previous implementation closed the launcher
        // at the start of a GUI program.
        // else
        //   PostMessage( hwnd, WM_CLOSE, 0, 0 );

        break;
        // TODO: test against specific error codes
      }
      case L'C': {
        STARTUPINFO si;
        PROCESS_INFORMATION pi;
        wchar_t * cdir;
        ZeroMemory( &si, sizeof(si) );
        si.cb = sizeof(si);
        si.lpTitle = ctrl->display;
        ZeroMemory( &pi, sizeof(pi) );
        cdir = get_shell_folder( FOLDERID_Documents ); // NULL if failure
        if( !CreateProcess( NULL, ctrl->action, NULL, NULL, TRUE, 0,
            NULL, cdir, &si, &pi ))
          tlwarn( hwnd, L"Failure to start %ls", ctrl->action );
        // else
        //   PostMessage( hwnd, WM_CLOSE, 0, 0 );
        break;
      }
      case L'P': {
        STARTUPINFO si;
        PROCESS_INFORMATION pi;
        wchar_t * cdir;
        ZeroMemory( &si, sizeof(si) );
        si.cb = sizeof(si);
        si.lpTitle = ctrl->display;
        ZeroMemory( &pi, sizeof(pi) );
        cdir = get_shell_folder( FOLDERID_Documents ); // NULL if failure
        if( !CreateProcess( NULL, ctrl->progid->command, NULL, NULL, TRUE, 0,
            NULL, cdir, &si, &pi ))
          tlwarn( hwnd, L"Failure to start %ls", ctrl->progid->command );
        // else
        //   PostMessage( hwnd, WM_CLOSE, 0, 0 );
        break;
      }
      case L'F': {
        if( ctrl->fun ) {
          ( ctrl->fun )( hwnd );
          break;
        } else {
          return FALSE;
        }
      }
      default: tldie( hwnd, L"Internal error" );
    } // switch
  } // ctrl->enabled
  return TRUE; // message handled
} // run_control

void get_icon( Control * ctrl ) {
  wchar_t * iconfile;
  UINT iconindex;
  if( !ctrl->type == L'P' )
    tldie( NULL, L"Get_icon invoked for wrong control %ls", ctrl->display );
  Progid * pgid = ctrl->progid;
  if( ctrl->hicon && !ctrl->sharedicon ) DestroyIcon( ctrl->hicon );
  ctrl->sharedicon = FALSE;
  if( pgid->command ) {
    if( pgid->iconspec )
      // while parsing the ini file, an iconspec was constructed
      // from the executable path name if not explicitly specified
      parse_iconspec( pgid->iconspec, &iconfile, &iconindex );
    if( iconfile) {
      ctrl->hicon = ExtractIcon( hInst, iconfile, iconindex );
      FREE0( iconfile );
    }
    if( !ctrl->hicon ) {
      ctrl->sharedicon = TRUE;
      ctrl->hicon = LoadIcon( NULL, MAKEINTRESOURCE( IDI_APPLICATION ));
    }
  } else {
    ctrl->sharedicon = TRUE;
    ctrl->hicon = LoadIcon( NULL, MAKEINTRESOURCE( IDI_ERROR ));
  }
} // get_icon

// used both for button captions and an optional announcement text
HWND create_label( HWND hparent, UINT_PTR id, wchar_t * display,
    int x_coord, int y_coord, int x_size, int y_size ) {
  HWND lbl;
  lbl = CreateWindow(
    L"STATIC",
    NULL,
    WS_CHILD | WS_VISIBLE | SS_CENTER,
    x_coord,
    y_coord,
    x_size,
    y_size,
    hparent,
    ( HMENU )id,
    hInst,
    NULL
  );
  if( !lbl ) tldie( NULL, L"Failure to create text label");
  SendMessage( lbl, WM_SETFONT, ( WPARAM )hfnt, TRUE );
  SendMessage( lbl, WM_SETTEXT, 0, (LPARAM)display );
  return lbl;
} // create_label

// buttons aligned horizontally;
// buttons_width is initially horizontal position of its left edge,
// afterwards left edge of next button

HWND create_text_button( HWND hparent, UINT_PTR id, wchar_t * display,
    int * buttons_width, int y_coord ) {
  int width, height = 0;
  HWND btn;
  width = ( wcslen( display ) + 4 ) * cx;
  height = 7 * cy / 4;
  btn = CreateWindow(
    L"BUTTON",
    display,
    WS_CHILD | WS_VISIBLE | WS_TABSTOP | BS_PUSHBUTTON,
    *buttons_width,
    y_coord,
    width,
    height,
    hparent,
    ( HMENU )id,
    hInst,
    NULL
  );
  if( !btn ) tldie( NULL, L"Failure to create text button" );
  SendMessage( btn, WM_SETFONT, ( WPARAM )hfnt, TRUE );
  *buttons_width += ( width + 4 * cx );
  return btn;
} // create_text_button

// icon buttons: arrange for a fixed-width text label centered underneath.
// buttons_width is updated only by create_button_caption.

HWND create_icon_button( HWND hparent, UINT_PTR id,
    HICON hicon, int * buttons_width, int y_coord ) {
  HWND btn;
  int lift = iconside - 7 * cy / 4;
  btn = CreateWindow(
      L"BUTTON",
      NULL,
      WS_CHILD | WS_VISIBLE | WS_TABSTOP | BS_PUSHBUTTON | BS_ICON,
      *buttons_width + ( labelmax * cx - iconside ) / 2,
      y_coord - lift,
      iconside,
      iconside,
      hparent,
      ( HMENU )id,
      hInst,
      NULL
  );
  if( !btn ) tldie( NULL, L"Failure to create icon button" );
  if( hicon )
      SendMessage( btn, BM_SETIMAGE, ( WPARAM )IMAGE_ICON, ( LPARAM )hicon );
  // otherwise leave button face blank;
  // *buttons_width not updated here
  return btn;
} // create_icon_button

HWND create_button_caption( HWND hparent, UINT_PTR id, wchar_t * display,
    int * buttons_width, int y_coord ) {
  HWND lbl;
  lbl = CreateWindow(
      L"STATIC",
      NULL,
      WS_CHILD | WS_VISIBLE | SS_CENTER,
      *buttons_width,
      y_coord + cy / 6,
      labelmax * cx,
      5 * cy / 2,
      hparent,
      ( HMENU )id,
      hInst,
      NULL
  );
  if( !lbl ) tldie( NULL, L"Failure to create text label" );
  SendMessage( lbl, WM_SETFONT, ( WPARAM )hfnt, TRUE );
  SendMessage( lbl, WM_SETTEXT, 0, (LPARAM)display );
  *buttons_width += ( labelmax + 4 ) * cx ;
  return lbl;
} // create_button_caption

void resize_toplevel( HWND hwnd, int w, int h, UINT flags ) {
  RECT r_cl, r_wn;
    GetClientRect( hwnd, &r_cl );
    GetWindowRect( hwnd, &r_wn );
    // adjust size for difference  between window size and client size.
    // top left client coordinate is always 0,0
    SetWindowPos(
        hwnd, HWND_TOP, 0, 0,
        w + r_wn.right - r_wn.left - r_cl.right,
        h + r_wn.bottom - r_wn.top - r_cl.bottom,
        flags );
} // resize_toplevel

// existence checking is done by parse_ini
void build_gui( HWND hwnd ) {
  int i, j, bwidth, mwidth, win_width, y;
  HMENU hMenu, hSubMenu;
  // global variable UINT_PTR id_next_control
  // is set to ID_FIRST_CONTROL in wWinMain.
  // tlauncher-specific structs:
  SubMenu * smenu;
  Control * ctrl, * gctrl;
  BOOL ingroup;
  int gleft;
  HWND lbl;

  SendMessage( hwnd, WM_SETTEXT, 0, (LPARAM)tl_name );
  if( is_uninstaller ) {
    bwidth = 4 * cx;
    y = 2 * cy;
    create_text_button( hwnd, ID_DONT, L"Abort", &bwidth, y );
    create_text_button( hwnd, ID_DO, L"Remove TeX Live settings",
        &bwidth, y );

    resize_toplevel( hwnd, bwidth, 5 * cy,
        SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOZORDER | SWP_NOREDRAW );
  } else { // ! is_uninstaller
    hMenu = CreateMenu();
    mwidth = 4 * cx;
    for( i=0; i<nmenus; i++ ) {
      smenu = menus[i];
      BOOL smenu_enabled = FALSE;
      if( !smenu ) continue; // should not happen
      if( smenu->nitems<1 ) continue;
      hSubMenu = CreatePopupMenu();
      for( j=0; j<smenu->nitems; j++ ) {
        ctrl = controls[smenu->first_item + j];
        if( ctrl->type == L'N' ) { // separator
          AppendMenu( hSubMenu, MF_SEPARATOR, 0, NULL );
        } else {
          wchar_t * s;
          ctrl->id = id_next_control++;
          if( ctrl == ed_control && ed_progid->name ) {
            s = def_editorlabel( FALSE );
            AppendMenu( hSubMenu, MF_STRING | MF_POPUP |
                ( ctrl->enabled ? MF_ENABLED : ( MF_DISABLED | MF_GRAYED )),
                ctrl->id, s );
            // this is the only possible disabled menu item.
            // other disabled controls will simply be skipped.
            FREE0( s );
            smenu_enabled = TRUE;
          } else if( ctrl->enabled ) {
            AppendMenu( hSubMenu, MF_STRING | MF_POPUP |
                ( ctrl->enabled ? MF_ENABLED : ( MF_DISABLED | MF_GRAYED )),
                ctrl->id, ctrl->display );
            smenu_enabled = TRUE;
          }
        }
      }
      if( smenu_enabled ) {
        AppendMenu( hMenu, MF_STRING | MF_POPUP |
            ( smenu_enabled ? MF_ENABLED : ( MF_DISABLED | MF_GRAYED )),
            (UINT_PTR)hSubMenu, smenu->name );
        mwidth = mwidth + cx * ( wcslen( smenu->name ) + 4 );
      }
    }
    SetMenu( hwnd, hMenu );
    // buttons: icon button + label for progids, text button for anything else
    // grouping of buttons is indicated by a surrounding rectangle
    bwidth = 2 * cx;
    if( nbuttons > 0 ) {
      y = ( tl_announce ? 8 : 5 ) * cy / 2;
      ingroup = FALSE;
      for ( i=0; i<nbuttons; i++ ) {
        wchar_t * s;
        ctrl = controls[first_button + i];
        if( ctrl->type == L'N' ) {
          // just drawing a rectangle in the main window would have to be done
          // during the processing of WM_PAINT. Logistically, it is simpler
          // to create a static window here, in WM_CREATE,
          // during group open, and properly size it at group close
          if( !ingroup ) {
            ingroup = TRUE;
            gctrl = ctrl;
            // memorize horizontal position
            gleft = bwidth - cx / 2;
            bwidth += cx;
            // width parameter will be fixed at group close
            ctrl->hw = CreateWindow( L"STATIC", NULL, WS_CHILD | WS_VISIBLE |
                SS_ETCHEDFRAME, gleft, y - iconside + cy, cx, 7 * cy, hwnd,
                NULL, hInst, NULL );
            continue;
          } else {
            if( !SetWindowPos( gctrl->hw, NULL, gleft, y - iconside + cy,
                bwidth - 2 * cx - gleft, 7 * cy, SWP_NOZORDER ))
              tlwarn( NULL, L"Failure to resize, error %u", GetLastError( ));
            ShowWindow( gctrl->hw, SW_SHOWNORMAL );
            bwidth += cx;
            ingroup = FALSE;
            continue;
          }
        }
        // remaining cases: actual buttons
        ctrl->id = id_next_control++;
        if( ctrl->type == L'P' ) {
          Progid * pgid = ctrl->progid;
          if( !pgid ) tldie( hwnd, L"No progid for %ls", ctrl->display );
          get_icon( ctrl );
          ctrl->hw = create_icon_button( hwnd, ctrl->id, ctrl->hicon,
              &bwidth, y );
          if( !ctrl->hw ) tldie( NULL, L"Failure to create button");
          if( ctrl == ed_control && ed_progid->name ) {
            // somehow, update_ed_control fails to fix this label
            // so we stupidly code the label separately here
            s = def_editorlabel( TRUE );
            create_button_caption( hwnd, id_next_control++, s,
                &bwidth, y + 2 * cy );
            FREE0( s );
          } else {
            create_button_caption( hwnd, id_next_control++, ctrl->display,
                &bwidth, y + 2 * cy );
          }
        } else {
          ctrl->hw = create_text_button( hwnd, ctrl->id, ctrl->display,
              &bwidth, y );
          // NULL icon => blank button, no error
          if( !ctrl->hw ) tldie( NULL, L"Failure to create button");
        }
        if( !ctrl->enabled ) EnableWindow( ctrl->hw, FALSE );
      }
    } // nbuttons > 0
    win_width = ( bwidth > mwidth ) ? bwidth : mwidth;
    // announcement text, if any
    if( tl_announce ) {
      lbl = CreateWindow(
          L"STATIC",
          NULL,
          WS_CHILD | WS_VISIBLE | SS_CENTER,
          cx,
          cy / 2,
          win_width - 2 * cx,
          3 * cy / 2,
          hwnd,
          ( HMENU )id_next_control++,
          hInst,
          NULL
      );
      if( lbl ) {
        SendMessage( lbl, WM_SETFONT, ( WPARAM )hfnt, TRUE );
        SendMessage( lbl, WM_SETTEXT, 0, (LPARAM)tl_announce );
      }
    }
    if( ed_control ) update_ed_control();
    // window height
    y = 5 * cy / 2;
    if( tl_announce ) y += 3 * cy / 2;
    if( nbuttons > 0 ) y += iconside + 3 * cy;
    resize_toplevel( hwnd, win_width, y,
        SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOZORDER | SWP_NOREDRAW );
  }
} // build_gui

LRESULT CALLBACK WndProc(HWND hwnd, UINT Mess, WPARAM wParam, LPARAM lParam) {
  switch (Mess) {
    case WM_CREATE: {
      HDC hdc;
      TEXTMETRIC tm;
      LOGFONT lf;

      InitCommonControls(); // needed for buttons with bitmaps

      // gui stuff
      hdc = GetDC( hwnd ); // NULL: DC for the entire screen

      // default font
      hfnt = GetStockObject( DEFAULT_GUI_FONT );
      SelectObject( hdc, hfnt );
      GetTextMetrics( hdc, &tm );
      // cx, cy, cxb and cyb are declared in tlaunch.h
      cx = tm.tmAveCharWidth;
      cy = tm.tmHeight + tm.tmExternalLeading;
      // big font, for splash screen
      hfntbig = NULL; //should already be true
      ZeroMemory( &lf, sizeof( lf ));
      if( GetObject( hfnt, sizeof( lf ), &lf )) {
        lf.lfHeight = 1.5 * lf.lfHeight;
        lf.lfWidth = 0;
        lf.lfPitchAndFamily = VARIABLE_PITCH | FF_SWISS;
        hfntbig = CreateFontIndirect( &lf );
      }
      if( hfntbig ) {
        SelectObject( hdc, hfntbig );
        GetTextMetrics( hdc, &tm );
        cxb = tm.tmAveCharWidth;
        cyb = tm.tmHeight + tm.tmExternalLeading;
      } else {
        hfntbig = hfnt;
        cxb = cx;
        cyb = cy;
      }
      // re-select stock font
      SelectObject( hdc, hfnt );

      // white background
      hwhite = ( HBRUSH )GetStockObject( WHITE_BRUSH );

      ReleaseDC( NULL, hdc );

      // start of resource numbering for menu items and push buttons
      id_next_control = ID_FIRST_CONTROL;
      // same for other child controls
      id_next_child = ID_FIRST_CHILD;

      get_config( hwnd ); // does its own error handling
      build_gui( hwnd );
      return 0;
    }
    case WM_CTLCOLORSTATIC: {
      HDC hdcStatic = ( HDC )wParam;
      // it makes no sense to use both an RGB value
      // and a stock brush, but that is what everybody tells me to do.
      // at least, it appears to work now (knock on wood)
      SetBkColor(hdcStatic, RGB(255,255,255));
      return (INT_PTR) hwhite; // global pointer to stock white brush
    }
    case WM_COMMAND: {
      // uninstaller buttons
      if( LOWORD(wParam) == ID_DONT ) {
        PostMessage( hwnd, WM_CLOSE, 0, 0 );
        return 0;
      } else if( LOWORD(wParam) == ID_DO ) {
        uninstall_self();
        PostMessage( hwnd, WM_CLOSE, 0, 0 );
        return 0;
      // launcher buttons
      } else if( LOWORD(wParam) >= ID_FIRST_CONTROL &&
          LOWORD(wParam) < id_next_control ) {
        // menu item or button
        if( run_control( hwnd, LOWORD(wParam) )) return 0;
        else return 1; // i.e. not handled by a control of ours
      }
      break;
    }
    case WM_CLOSE:
      DestroyWindow(hwnd);
      return 0;
    case WM_DESTROY:
      PostQuitMessage(0);
      return 0;
    default:
      return DefWindowProc(hwnd, Mess, wParam, lParam);
  }
  return 1; // makes the compiler happy; we should not get here at all
}

// boilerplate

int WINAPI wWinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
    PWSTR szCmdLine, int iCmdShow) {

  wchar_t * mainwin_class;
  WNDCLASSEX wclass;
  HWND hDup; // hMain is already global
  MSG Msg;

  DWORD pid, sid;
  wchar_t * username;

  set_simple_globals();

  // Avoid running multiple copies per user session:
  // compute name of main window class from username and session id,
  // then check for other occurrences of this window class.
  // Since this may be a restart, wait a moment for any parent to shut down.
  username = get_env( L"USERNAME" );
  pid = GetCurrentProcessId();
  if( username && ProcessIdToSessionId( pid, &sid )) { // otherwise dont check
    tl_wasprintf( &mainwin_class, L"tlaunch-%ls%010lu", username, sid );
    Sleep( 1000 );
    hDup = FindWindow( mainwin_class, NULL );
    if( hDup ) {
      ShowWindow( hDup, SW_RESTORE );
      if( !SetForegroundWindow( hDup )) tlinfo( NULL, L"Already running" );
      ExitProcess( 0 );
    }
  }
  FREE0( username );

  hInst = hInstance; // copy to global variable

  wclass.cbSize     = sizeof(WNDCLASSEX);
  wclass.style     = CS_HREDRAW | CS_VREDRAW;
  wclass.lpfnWndProc   = WndProc;
  wclass.cbClsExtra   = 0;
  wclass.cbWndExtra   = 0;
  wclass.hInstance   = hInstance;
  wclass.hIcon       = LoadIcon( GetModuleHandle( NULL ), MAKEINTRESOURCE( IDI_TL ));
  wclass.hCursor     = LoadCursor(NULL, IDC_ARROW);
  wclass.hbrBackground = (HBRUSH)GetStockObject (WHITE_BRUSH);
  wclass.lpszMenuName  = NULL;
  wclass.lpszClassName = mainwin_class;
  wclass.hIconSm     = NULL;

  if(!RegisterClassEx(&wclass)) {
    tldie( NULL, L"Window Registration Failed!" );
    return 0;
  }

  // splash screen for long-running scripts
  register_splash( );

  // window for options dialog
  register_opt_wnd_class( );

  hMain = CreateWindow( mainwin_class, L"Launcher",
      WS_OVERLAPPEDWINDOW,
      CW_USEDEFAULT, CW_USEDEFAULT,
      640, 480, // width and height to be reset during WM_CREATE callback
      NULL, NULL, hInst, NULL );
  // The menu will be hooked up to the main window in build_gui;
  // for now it is still NULL
  if( !hMain ) {
    tldie( NULL, L"Failure to create main window" );
  }
  ShowWindow(hMain, iCmdShow);
  UpdateWindow(hMain);

  while(GetMessage(&Msg, NULL, 0, 0) > 0) {
    TranslateMessage(&Msg);
    DispatchMessage(&Msg);
  }
  return Msg.wParam;
}
